#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>


class TobiiJSONProperty:

    def __init__(self, json_key, py_type):
        self.key = json_key
        self.py_type = py_type

    def asValue(self, value):
        if type(self.py_type) is list:
            ret = []
            for i in range(0, len(value)):
                ret.append( self.py_type[i](value[i]) )
        else:
            ret = self.py_type(value)
        return ret

class TobiiJSONProperties:

    APISynch_ETS = TobiiJSONProperty("ets", str)
    APISynch_TAG = TobiiJSONProperty("tag", str)
    APISynch_Type = TobiiJSONProperty("type", str)
    Eye = TobiiJSONProperty("eye", str)
    GazeIndex = TobiiJSONProperty("gidx", int)
    GazeDirection = TobiiJSONProperty("gd", [float, float, float])
    GazePosition = TobiiJSONProperty("gp", [float, float])
    GazePosition3d = TobiiJSONProperty("gp3", [float, float, float])
    MEMS_Accelerometer = TobiiJSONProperty("acc", [float, float, float])
    MEMS_Gyroscope = TobiiJSONProperty("gy", [float, float, float])
    PupilCenter = TobiiJSONProperty("pc", [float, float, float])
    PupilDiameter = TobiiJSONProperty("pd", float)
    PTS = TobiiJSONProperty("pts", int)
    PV = TobiiJSONProperty("pv", int)
    Status = TobiiJSONProperty("s", int)
    SynchPort_Dir = TobiiJSONProperty("dir", str)
    SynchPort_Sig = TobiiJSONProperty("sig", str)
    Timestamp = TobiiJSONProperty("ts", int)
    VTS = TobiiJSONProperty("vts", int)

class TobiiJSONAttribute:

    def __init__(self, tobii_property, json_sample):
        self.__tobii_property__ = tobii_property
        self.__value__ = self.__tobii_property__.asValue(json_sample[self.__tobii_property__.key])

    def getKey(self):
        return self.__tobii_property__.key

    def getType(self):
        return self.__tobii_property__.py_type

    def getValue(self):
        return self.__value__


class PupilCenter:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)
        self.gidx = TobiiJSONAttribute(TobiiJSONProperties.GazeIndex, json_sample)
        self.pc = TobiiJSONAttribute(TobiiJSONProperties.PupilCenter, json_sample)
        self.eye = TobiiJSONAttribute(TobiiJSONProperties.Eye, json_sample)


class PupilDiameter:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)
        self.gidx = TobiiJSONAttribute(TobiiJSONProperties.GazeIndex, json_sample)
        self.pd = TobiiJSONAttribute(TobiiJSONProperties.PupilDiameter, json_sample)
        self.eye = TobiiJSONAttribute(TobiiJSONProperties.Eye, json_sample)

class GazeDirection:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)
        self.gidx = TobiiJSONAttribute(TobiiJSONProperties.GazeIndex, json_sample)
        self.gd = TobiiJSONAttribute(TobiiJSONProperties.GazeDirection, json_sample)
        self.eye = TobiiJSONAttribute(TobiiJSONProperties.Eye, json_sample)

class GazePosition:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)
        self.gidx = TobiiJSONAttribute(TobiiJSONProperties.GazeIndex, json_sample)
        self.gp = TobiiJSONAttribute(TobiiJSONProperties.GazePosition, json_sample)


class GazePosition3d:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)
        self.gidx = TobiiJSONAttribute(TobiiJSONProperties.GazeIndex, json_sample)
        self.gp3 = TobiiJSONAttribute(TobiiJSONProperties.GazePosition3d, json_sample)

class APISynch:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)
        self.ets = TobiiJSONAttribute(TobiiJSONProperties.APISynch_ETS, json_sample)
        self.type = TobiiJSONAttribute(TobiiJSONProperties.APISynch_Type, json_sample)
        self.tag = TobiiJSONAttribute(TobiiJSONProperties.APISynch_TAG, json_sample)

class VTS:

    def __init__(self, json_sample):
        self.ts = TobiiJSONAttribute(TobiiJSONProperties.Timestamp, json_sample)
        self.vts = TobiiJSONAttribute(TobiiJSONProperties.VTS, json_sample)
        self.s = TobiiJSONAttribute(TobiiJSONProperties.Status, json_sample)


#1. To define other packages from C.6.1.6 to  C.6.1.13



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import datetime
import logging
import os
import tobiiglasses.utils as utils

TOBII_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S+%f'
PROJECTS_DIRNAME = "projects"
RECORDINGS_DIRNAME = "recordings"
SEGMENTS_DIRNAME = "segments"

def get_recordings_path(root_dir, project_id):
    return os.path.join(root_dir, PROJECTS_DIRNAME, project_id, RECORDINGS_DIRNAME)

def get_projects_path(root_dir):
    return os.path.join(root_dir, PROJECTS_DIRNAME)

def get_all_projects(root_dir):
    logging.info('Loading projects in %s' % root_dir)
    projects = []
    projects_path = os.path.join(root_dir, PROJECTS_DIRNAME)
    for item in os.listdir (projects_path):
        if os.path.isdir(os.path.join(projects_path, item)):
            projects.append(TobiiProject(projects_path, item))
    return projects

def get_all_recordings(root_dir, project_id):
    recordings_path = get_recordings_path(root_dir, project_id)
    logging.info('Loading recordings from %s' % recordings_path)
    recordings = []
    for item in os.listdir (recordings_path):
        if os.path.isdir(os.path.join(recordings_path, item)):
            recordings.append(TobiiRecording(recordings_path, item))
    return recordings

def get_all_segments(recordings_path, recording_id):
    segments_path = os.path.join(recordings_path, recording_id, SEGMENTS_DIRNAME)
    logging.info('Loading segments from %s' % segments_path)
    segments = []
    for item in os.listdir (segments_path):
        seg_path = os.path.join(segments_path, item)
        if os.path.isdir(seg_path):
            segments.append(TobiiSegment(seg_path))
    return segments

class TobiiRecording:
    rec_filename = "recording.json"

    def __init__(self, recordings_path, rec_id):
        self.__rec_id__ = rec_id
        self.__rec_path__ = os.path.join(recordings_path, rec_id)
        self.__project_dir__ = os.path.dirname(os.path.dirname(os.path.abspath(self.__rec_path__)))
        res = utils.load_json_from_file(self.__rec_path__, TobiiRecording.rec_filename)
        self.__parse_json_rec__(res)
        self.__project__ = TobiiProject(os.path.dirname(self.__project_dir__), os.path.basename(self.__project_dir__))
        self.__participant__ = TobiiParticipant(self.__rec_path__)
        self.__segments__ = get_all_segments(recordings_path, self.__rec_id__)

    def __parse_json_rec__(self, item):
        self.__rec_created__ = datetime.datetime.strptime(item["rec_created"], TOBII_DATETIME_FORMAT)
        self.__rec_name__ = item["rec_info"]["Name"]
        self.__rec_length__ = int(item["rec_length"])
        self.__rec_segments__ = int(item["rec_segments"])
        self.__rec_et_samples__ = int(item["rec_et_samples"])
        self.__rec_et_valid_samples__ = int(item["rec_et_valid_samples"])

    def getAttributes(self):
        attr = []
        names = []
        names.append("Recording name"); attr.append(self.__rec_name__)
        names.append("Project name"); attr.append(self.__project__.getName())
        names.append("Creation Date"); attr.append(str(self.__rec_created__))
        names.append("Duration (s)"); attr.append(str(self.__rec_length__))
        names.append("Participant name"); attr.append(self.__participant__.getName())
        names.append("Segments nr"); attr.append(str(self.__rec_segments__))
        names.append("Et samples"); attr.append(str(self.__rec_et_samples__))
        names.append("Et valid samples"); attr.append(str(self.__rec_et_valid_samples__))
        return (names, attr)

    def getCreationDate(self):
        return self.__rec_created__

    def getEtSamples(self):
        return self.__rec_et_samples__

    def getEtValidSamples(self):
        return self.__rec_et_valid_samples__

    def getId(self):
        return self.__rec_id__

    def getLength(self):
        return self.__rec_length__

    def getName(self):
        return self.__rec_name__

    def getParticipant(self):
        return self.__participant__

    def getRecordingDir(self):
        return self.__rec_dir__

    def getSegment(self, segment_id):
        if segment_id > 0:
            return self.__segments__[segment_id-1]
        logging.error('Cannot get segment less or equal to zero')
        return None

    def getSegmentsN(self):
        return self.__rec_segments__


class TobiiProject:
    project_filename = "project.json"

    def __init__(self, projects_path, project_id):
        self.__project_id__ = project_id
        self.__project_path__ = os.path.join(projects_path, project_id)
        res = utils.load_json_from_file(self.__project_path__, TobiiProject.project_filename)
        self.__parse_json_pr__(res)

    def __parse_json_pr__(self, item):
        self.__pr_created__ = datetime.datetime.strptime(item["pr_created"], TOBII_DATETIME_FORMAT)
        try:
            self.__pr_name__ = item["pr_info"]["Name"]
        except:
            self.__pr_name__ = None

    def getCreationDate(self):
        return self.__pr_created__

    def getId(self):
        return self.__project_id__

    def getName(self):
        return self.__pr_name__



class TobiiParticipant:
    ppt_filename = "participant.json"

    def __init__(self, ppt_path):
        res = utils.load_json_from_file(ppt_path, TobiiParticipant.ppt_filename)
        self.__parse_json_ppt__(res)

    def __parse_json_ppt__(self, item):
        self.__ppt_name__ = item["pa_info"]["Name"]

    def getName(self):
        return self.__ppt_name__

class TobiiSegment:
    video_filename = "fullstream.mp4"
    livedata_filename = "livedata.json.gz"
    segment_filename = "segment.json"

    def __init__(self, segment_path):
        self.__segment_path__ = segment_path
        self.__segment_id__ = os.path.basename(segment_path)
        res = utils.load_json_from_file(segment_path, TobiiSegment.segment_filename)
        if res != None:
            self.__parse_json_seg__(res)

    def __parse_json_seg__(self, item):
        self.__seg_length_us__ = int(item["seg_length_us"])
        self.__seg_calibrated__ = bool(item["seg_calibrated"])
        self.__seg_t_start__ = datetime.datetime.strptime(item["seg_t_start"], TOBII_DATETIME_FORMAT)
        self.__seg_t_stop__ = datetime.datetime.strptime(item["seg_t_stop"], TOBII_DATETIME_FORMAT)

    def getId(self):
        return self.__segment_id__

    def getLengthUs(self):
        return self.__seg_length_us__

    def getLiveDataFilepath(self):
        return os.path.join(self.__segment_path__, self.__segment_id__)

    def getSegmentPath(self):
        return self.__segment_path__

    def getStartDateTime(self):
        return self.__seg_t_start__

    def getStopDateTime(self):
        return self.__seg_t_stop__

    def getVideoFilename(self):
        return os.path.join(self.__segment_path__, TobiiSegment.video_filename)

    def isCalibrated(self):
        return self.__seg_calibrated__



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import itertools
import numpy as np
import joblib
from scipy import linalg
from scipy.misc import imread
import matplotlib.pyplot as plt
import matplotlib as mpl

from sklearn import mixture
from sklearn.cluster import DBSCAN as DB

class AOI_Model:
    COLORS_LIST = ['red', 'orange', 'yellow', 'green', 'cyan', 'blue', 'magenta']

    def __init__(self, n_clusters=None, cluster_labels=[], sorted_key=None):
        self.__n_clusters__ = n_clusters
        self.__cluster_labels__ = cluster_labels
        self.__sorted_key__ = sorted_key
        self.__initvars__()

    def __initvars__(self):
        self.__model__ = None
        self.__means__ = None
        self.__ordered_means__ = None
        self.__covariances__ = None
        self.__X__ = None
        self.__labels__ = None

    def __indexOf__(self, ordered_means, value):
        i = 0
        for m in ordered_means:
            if value[0] == ordered_means[i][0] and ordered_means[i][1]:
                return i
            i+=1
        return -1

    def __assignColorLabels__(self, labels):
        newlabels = []
        for l in labels:
            newlabels.append(AOI_Model.COLORS_LIST[l])
        return newlabels

    def __assignsLabels__(self, means, labels, cluster_labels, sorted_key):
        if not self.__sorted_key__  is None and len(self.__cluster_labels__) != self.__n_clusters__:
            raise ValueError('The AOI filter requires an equal numbers of clusters and labels')
        self.__ordered_means__ = []
        for m in sorted(means, key=sorted_key):
            self.__ordered_means__.append( m.astype(np.float16))
        new_labels = []
        for item in labels:
            m = means[item].astype(np.float16)
            new_labels.append(cluster_labels[self.__indexOf__(self.__ordered_means__, m)])
        return new_labels

    def __color_iter__(self):
        return itertools.cycle(AOI_Model.COLORS_LIST)

    def __draw_ellipse__(self, mean, covar, color, cplot):
        v, w = linalg.eigh(covar)
        v = 2. * np.sqrt(2.) * np.sqrt(v)
        u = w[0] / linalg.norm(w[0])
        angle = np.arctan(u[1] / u[0])
        angle = 180. * angle / np.pi  # convert to degrees
        ell = mpl.patches.Ellipse(mean, v[0], v[1], 180. + angle, color=color)
        ell.set_clip_box(cplot.bbox)
        ell.set_alpha(0.5)
        cplot.add_artist(ell)

    def __fit__(self, X):
        """
        to assign:
        self.__X__, self.__n_clusters__, self.__labels__, self.__means__, self.__covariances__
        """
        raise NotImplementedError( "AOIs should have implemented a fit function" )

    def __getAOI_Distance__(self, fixation_x, fixation_y, cluster_mean):
        c = np.array((cluster_mean[0], cluster_mean[1]))
        x = np.array((fixation_x, fixation_y))
        return np.linalg.norm(x-c)

    def getCentroid(self, aoi_label):
        return (self.__means__[aoi_label][0], self.__means__[aoi_label][1])

    def getEllipseParams(self, aoi_label):
        v, w = linalg.eigh(self.__covariances__[aoi_label])
        v = 2. * np.sqrt(2.) * np.sqrt(v)
        u = w[0] / linalg.norm(w[0])
        angle = np.arctan(u[1] / u[0])
        angle = 180. * angle / np.pi  # convert to degrees
        return (v[0], v[1], angle+180.0)

    def fit(self, gaze_events, ts_filter):
        self.__initvars__()
        ts_list, x, y = gaze_events.getFixationsAsNumpy(ts_filter)
        self.__X__ = np.column_stack((x,y))
        (self.__model__,self.__means__, self.__covariances__, self.__labels__) = self.__fit__(self.__X__)
        #self.__scores__ = self.__model__.score_samples(self.__X__)
        if not self.__sorted_key__ is None:
            self.__labels__ = self.__assignsLabels__(self.__means__, self.__labels__, self.__cluster_labels__, self.__sorted_key__)
        color_labels = self.__assignColorLabels__(self.__labels__)
        i = 0
        for ts in ts_list:
            gaze_events.setAOI(ts, color_labels[i], self.__getAOI_Distance__(self.__X__[i][0], self.__X__[i][1], self.__means__[self.__labels__[i]]))
            i+=1

    def saveModel(self, filename):
        joblib.dump(self.__model__, filename)

    def savePlot(self, title='no title', filename='plot.pdf', background_image=None, width=1920, height=1080):
        p = self.__plot__(title, background_image, width, height)
        p.savefig(filename, format='pdf')
        p.gca().clear()

    def showPlot(self, title='no title', background_image=None, width=1920, height=1080):
        p = self.__plot__(title, background_image, width, height)
        p.show()
        p.gca().clear()

    def __plot__(self, title, background_image=None, width=1920, height=1080):
        cplot = plt.gca()
        img = None
        if not background_image is None:
            img = imread(background_image)
        for i, (mean, covar, color) in enumerate(zip(self.__means__, self.__covariances__, self.__color_iter__())):
            mean = list(map(np.float16, mean))
            if not np.any(self.__labels__ == i):
                continue
            plt.scatter(self.__X__[self.__labels__ == i, 0], self.__X__[self.__labels__ == i, 1], 1 , color=color, zorder=1)
            self.__draw_ellipse__(mean, covar, color, cplot)
            if not self.__ordered_means__ is None:
                plt.text(mean[0], mean[1], self.__cluster_labels__[self.__indexOf__(self.__ordered_means__, mean)], fontsize=12, bbox=dict(facecolor='red', alpha=0.5))
        plt.xlim(0, width)
        plt.ylim(0, height)
        plt.gca().invert_yaxis()
        plt.xticks(())
        plt.yticks(())
        plt.title(title)
        if not background_image is None:
            plt.imshow(img, zorder=0)
        return plt


class GaussianMixture(AOI_Model):

    def __init__(self, n_clusters, cluster_labels=[], sorted_key=None):
        AOI_Model.__init__(self, n_clusters, cluster_labels, sorted_key)

    def __fit__(self, X):
        model = mixture.GaussianMixture(n_components=self.__n_clusters__, covariance_type='full').fit(X)
        labels = model.predict(X)
        proba = model.predict_proba(X)
        return (model, model.means_, model.covariances_, labels)


class DBSCAN(AOI_Model):

    def __init__(self, eps, min_samples):
        AOI_Model.__init__(self)
        self.__eps__ = eps
        self.__min_samples__ = min_samples

    def __fit__(self, X):
        model = DB(eps=self.__eps__, min_samples=self.__min_samples__).fit(X)
        core_samples_mask = np.zeros_like(model.labels_, dtype=bool)
        core_samples_mask[model.core_sample_indices_] = True
        labels = model.labels_
        unique_labels = set(labels)
        n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
        self.__n_clusters__ = n_clusters_
        covariances = []
        means = []
        for label in unique_labels:
            if label == -1:
                continue
            class_member_mask = (labels == label)
            xy = X[class_member_mask & core_samples_mask]
            covariances.append(np.cov(np.stack((xy[:,0], xy[:,1]))))
            means.append([np.mean(xy[:,0]), np.mean(xy[:,1])])
        return (model, means, covariances, labels)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import logging
import os
import pandas as pd
import tobiiglasses as TG
from tobiiglasses.gazedata import GazeData



class CSVFile(object):

    def __init__(self, filepath, filename):
        self.__filepath__ = filepath
        self.__filename__ = filename
        self.__headers__ = []

    def __exportDataFrame__(self, df):
        dest = os.path.join(self.__filepath__, self.__filename__)
        logging.info('Exporting data in %s' % dest)
        df.to_csv(dest, encoding='utf-8', header=True, columns=self.__headers__, index=False)

    def getHeaders(self):
        return self.__headers__

    def setHeaders(self):
        return self.__headers__

    def toCSV(self):
        raise NotImplementedError( "CSVFile should have implemented toCSV method" )


class RawCSV(CSVFile):

    def __init__(self, filepath, filename, gazedata):
        CSVFile.__init__(self, filepath, filename)
        self.__gazedata__ = gazedata
        self.__headers__.append(GazeData.Timestamp)
        self.__headers__.append(GazeData.Gidx)
        self.__headers__.append(GazeData.LoggedEvents)
        self.__headers__.append(GazeData.GazePositionX)
        self.__headers__.append(GazeData.GazePositionY)
        self.__headers__.append(GazeData.GazePixelX)
        self.__headers__.append(GazeData.GazePixelY)
        self.__headers__.append(GazeData.Gaze3DPositionX)
        self.__headers__.append(GazeData.Gaze3DPositionY)
        self.__headers__.append(GazeData.Gaze3DPositionZ)
        self.__headers__.append(GazeData.Depth)
        self.__headers__.append(GazeData.Vergence)
        self.__headers__.append(GazeData.Version)
        self.__headers__.append(GazeData.Tilt)
        self.__headers__.append(GazeData.GazeDirectionX_Left)
        self.__headers__.append(GazeData.GazeDirectionY_Left)
        self.__headers__.append(GazeData.GazeDirectionZ_Left)
        self.__headers__.append(GazeData.GazeDirectionX_Right)
        self.__headers__.append(GazeData.GazeDirectionY_Right)
        self.__headers__.append(GazeData.GazeDirectionZ_Right)
        self.__headers__.append(GazeData.PupilCenterX_Left)
        self.__headers__.append(GazeData.PupilCenterY_Left)
        self.__headers__.append(GazeData.PupilCenterZ_Left)
        self.__headers__.append(GazeData.PupilCenterX_Right)
        self.__headers__.append(GazeData.PupilCenterY_Right)
        self.__headers__.append(GazeData.PupilCenterZ_Right)
        self.__headers__.append(GazeData.PupilDiameter_Left)
        self.__headers__.append(GazeData.PupilDiameter_Right)

    def toCSV(self):
        self.__exportDataFrame__(self.__gazedata__.toDataFrame())

class ExtendedRawCSV(RawCSV):

    def __init__(self, filepath, filename, gazedata):
        RawCSV.__init__(self, filepath, filename, gazedata)
        self.__headers__.extend(gazedata.getExpVarsHeaders())

class FixationsCSV(CSVFile):

    def __init__(self, filepath, filename, fixations_df):
        CSVFile.__init__(self, filepath, filename)
        self.__headers__.append(TG.events.GazeEvents.Timestamp)
        self.__headers__.append(TG.events.GazeEvents.EventIndex)
        self.__headers__.append(TG.events.GazeEvents.Fixation_X)
        self.__headers__.append(TG.events.GazeEvents.Fixation_Y)
        self.__headers__.append(TG.events.GazeEvents.EventDuration)
        self.__headers__.append(TG.events.GazeEvents.AOI)
        self.__headers__.append(TG.events.GazeEvents.AOI_Distance)
        self.__fixations_df__ = fixations_df

    def toCSV(self):
        self.__exportDataFrame__(self.__fixations_df__)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import logging
import math
import os
import re
import numpy as np
import pandas as pd
import tobiiglasses.entities
import tobiiglasses.utils
from sortedcontainers import SortedList, SortedDict
from tobiiglasses.livedata import *

logging.basicConfig(format='[%(levelname)s]: %(message)s', level=logging.DEBUG)

class GazeItem:
    def __init__(self, label, dtype):
        self.__label__ = label
        self.__data__ = SortedDict({})
        self.__dtype__ = dtype

    def __getitem__(self, key):
        return self.__data__[key]

    def __setitem__(self, key, value):
        self.__data__[key] = value

    def getData(self):
        return self.__data__

    def getLabel(self):
        return self.__label__

    def getType(self):
        return self.__dtype__

    def keys(self):
        return list(self.__data__.keys())

    def pop(self, key):
        self.__data__.pop(key)

    def values(self):
        return list(self.__data__.values())

class GazeData:
    Timestamp = "Timestamp"
    Gidx = "Gaze Index"
    LoggedEvents = "Logged Events"
    GazePositionX = "Gaze Position X"
    GazePositionY = "Gaze Position Y"
    Gaze3DPositionX = "Gaze 3D Position X"
    Gaze3DPositionY = "Gaze 3D Position Y"
    Gaze3DPositionZ = "Gaze 3D Position Z"
    Depth = "Gaze Depth"
    Tilt = "Gaze Tilt"
    Vergence = "Gaze Vergence"
    Version = "Gaze Version"
    GazeDirectionX_Left = "Gaze Direction Left X"
    GazeDirectionY_Left = "Gaze Direction Left Y"
    GazeDirectionZ_Left = "Gaze Direction Left Z"
    GazeDirectionX_Right = "Gaze Direction Right X"
    GazeDirectionY_Right = "Gaze Direction Right Y"
    GazeDirectionZ_Right = "Gaze Direction Right Z"
    GazePixelX = "Gaze Pixel X"
    GazePixelY = "Gaze Pixel Y"
    PupilCenterX_Left = "Pupil Center Left X"
    PupilCenterY_Left = "Pupil Center Left Y"
    PupilCenterZ_Left = "Pupil Center Left Z"
    PupilCenterX_Right = "Pupil Center Right X"
    PupilCenterY_Right = "Pupil Center Right Y"
    PupilCenterZ_Right = "Pupil Center Right Z"
    PupilDiameter_Left = "Pupil Diameter Left"
    PupilDiameter_Right = "Pupil Diameter Right"


    def __init__(self, segment):
        self.__gazedata__ = {}
        self.__gd_right__ = {}
        self.__gd_left__ = {}
        self.__experiment_vars__ = {} # keys: var_name, values: {keys: ts,  values: var_value}
        self.__experiment_vars_headers__ = []
        self.__rec_id__ = None
        self.__pa_name__ = None
        self.__pr_name__ = None
        self.__segment__ = segment
        self.__vts__ = SortedDict({})
        self.__init_datatypes__()
        self.__importGazeData__(segment)
        self.__importExpVars__()

    def __addItem__(self, livedatajson_item):
        if livedatajson_item is None:
            return
        if livedatajson_item.s.getValue() != 0: # Discard invalid samples
            return
        ts = livedatajson_item.ts.getValue()/1000.0

        if isinstance(livedatajson_item, VTS):
            self.__vts__[ts] = livedatajson_item.vts.getValue()/1000.0
            return
        else:
            self.__gazedata__[GazeData.Timestamp][ts] = ts # Conversion in milliseconds
        if isinstance(livedatajson_item, APISynch):
            tvalue = livedatajson_item.type.getValue()
            tag = livedatajson_item.tag.getValue()
            if tvalue != "JsonEvent":

                if tvalue.startswith('#') and tvalue.endswith('#'):
                    var_name = tvalue[1:-1]
                    if not var_name in self.__experiment_vars__.keys():
                        self.__experiment_vars__[var_name] = SortedDict({})
                    self.__experiment_vars__[var_name][ts] = tag

                elif tvalue.startswith('@') and tvalue.endswith('@'):
                    vars_list = eval(tvalue[1:-1])
                    values_list = eval(tag)
                    for i in range(0, len(vars_list)):
                        var_name = vars_list[i]
                        value = values_list[i]
                        if not var_name in self.__experiment_vars__.keys():
                            self.__experiment_vars__[var_name] = SortedDict({})
                        self.__experiment_vars__[var_name][ts] = value
                else:
                    self.__gazedata__[GazeData.LoggedEvents][ts] = tvalue

            else:
                self.__gazedata__[GazeData.Timestamp].pop(ts)
            return
        else:
            gidx = livedatajson_item.gidx.getValue()
            self.__gazedata__[GazeData.Gidx][ts] = gidx
        if isinstance(livedatajson_item, GazePosition):
            self.__gazedata__[GazeData.GazePositionX][ts] = livedatajson_item.gp.getValue()[0]
            self.__gazedata__[GazeData.GazePositionY][ts] = livedatajson_item.gp.getValue()[1]
            self.__gazedata__[GazeData.GazePixelX][ts] = int(1920*self.__gazedata__[GazeData.GazePositionX][ts])
            self.__gazedata__[GazeData.GazePixelY][ts] = int(1080*self.__gazedata__[GazeData.GazePositionY][ts])
            return
        elif isinstance(livedatajson_item, GazePosition3d):
            self.__gazedata__[GazeData.Gaze3DPositionX][ts] = livedatajson_item.gp3.getValue()[0]
            self.__gazedata__[GazeData.Gaze3DPositionY][ts] = livedatajson_item.gp3.getValue()[1]
            self.__gazedata__[GazeData.Gaze3DPositionZ][ts] = livedatajson_item.gp3.getValue()[2]
            self.__gazedata__[GazeData.Depth][ts] = (self.__gazedata__[GazeData.Gaze3DPositionX][ts]**2 + self.__gazedata__[GazeData.Gaze3DPositionY][ts]**2 + self.__gazedata__[GazeData.Gaze3DPositionZ][ts]**2)**0.5
            self.__gazedata__[GazeData.Tilt][ts] = math.atan(self.__gazedata__[GazeData.Gaze3DPositionY][ts]/self.__gazedata__[GazeData.Gaze3DPositionZ][ts])*180./math.pi
            return
        elif isinstance(livedatajson_item, GazeDirection):
            if livedatajson_item.eye.getValue() == "left":
                self.__gazedata__[GazeData.GazeDirectionX_Left][ts] = livedatajson_item.gd.getValue()[0]
                self.__gazedata__[GazeData.GazeDirectionY_Left][ts] = livedatajson_item.gd.getValue()[1]
                self.__gazedata__[GazeData.GazeDirectionZ_Left][ts] = livedatajson_item.gd.getValue()[2]
                self.__gazedata__[GazeData.Version][ts] = math.acos(self.__gazedata__[GazeData.GazeDirectionZ_Left][ts])*180./math.pi
                self.__gd_left__[gidx] = np.array([self.__gazedata__[GazeData.GazeDirectionX_Left][ts],self.__gazedata__[GazeData.GazeDirectionY_Left][ts], self.__gazedata__[GazeData.GazeDirectionZ_Left][ts]])
                return
            elif livedatajson_item.eye.getValue() == "right":
                self.__gazedata__[GazeData.GazeDirectionX_Right][ts] = livedatajson_item.gd.getValue()[0]
                self.__gazedata__[GazeData.GazeDirectionY_Right][ts] = livedatajson_item.gd.getValue()[1]
                self.__gazedata__[GazeData.GazeDirectionZ_Right][ts] = livedatajson_item.gd.getValue()[2]
                self.__gazedata__[GazeData.Version][ts] = math.acos(self.__gazedata__[GazeData.GazeDirectionZ_Right][ts])*180./math.pi
                self.__gd_right__[gidx] = np.array([self.__gazedata__[GazeData.GazeDirectionX_Right][ts], self.__gazedata__[GazeData.GazeDirectionY_Right][ts], self.__gazedata__[GazeData.GazeDirectionZ_Right][ts]])
            try:
                self.__gazedata__[GazeData.Vergence][ts] = math.acos( np.dot(self.__gd_left__[gidx], self.__gd_right__[gidx])/(np.linalg.norm(self.__gd_left__[gidx])*np.linalg.norm(self.__gd_right__[gidx])) )*180./math.pi
            except:
                pass
        elif isinstance(livedatajson_item, PupilCenter):
            if livedatajson_item.eye.getValue() == "left":
                self.__gazedata__[GazeData.PupilCenterX_Left][ts] = livedatajson_item.pc.getValue()[0]
                self.__gazedata__[GazeData.PupilCenterY_Left][ts] = livedatajson_item.pc.getValue()[1]
                self.__gazedata__[GazeData.PupilCenterZ_Left][ts] = livedatajson_item.pc.getValue()[2]
                return
            elif livedatajson_item.eye.getValue() == "right":
                self.__gazedata__[GazeData.PupilCenterX_Right][ts] = livedatajson_item.pc.getValue()[0]
                self.__gazedata__[GazeData.PupilCenterY_Right][ts] = livedatajson_item.pc.getValue()[1]
                self.__gazedata__[GazeData.PupilCenterZ_Right][ts] = livedatajson_item.pc.getValue()[2]
                return
        elif isinstance(livedatajson_item, PupilDiameter):
            if livedatajson_item.eye.getValue() == "left":
                self.__gazedata__[GazeData.PupilDiameter_Left][ts] = livedatajson_item.pd.getValue()
                return
            elif livedatajson_item.eye.getValue() == "right":
                self.__gazedata__[GazeData.PupilDiameter_Right][ts] = livedatajson_item.pd.getValue()
                return

    def __decodeJSON__(self, json_item):
        if TobiiJSONProperties.PupilCenter.key in json_item:
            item = PupilCenter(json_item)
        elif TobiiJSONProperties.PupilDiameter.key in json_item:
            item = PupilDiameter(json_item)
        elif TobiiJSONProperties.GazeDirection.key in json_item:
            item = GazeDirection(json_item)
        elif TobiiJSONProperties.GazePosition.key in json_item:
            item = GazePosition(json_item)
        elif TobiiJSONProperties.GazePosition3d.key in json_item:
            item = GazePosition3d(json_item)
        elif TobiiJSONProperties.APISynch_ETS.key in json_item:
            item = APISynch(json_item)
        elif TobiiJSONProperties.VTS.key in json_item:
            item = VTS(json_item)
        else:
            item = None
        self.__addItem__(item)

    def __importExpVars__(self):
        self.__experiment_vars_headers__ = []
        for k,v in self.__experiment_vars__.items():
            current_var = None
            self.__gazedata__[k] = GazeItem(k, np.dtype(str))
            self.__experiment_vars_headers__.append(k)
            for ts in self.__gazedata__[GazeData.Timestamp].getData().values():
                try:
                    current_var = v[ts]
                except:
                    pass
                self.__gazedata__[k][ts] = current_var

    def __importGazeData__(self, segment):
        filepath = segment.getSegmentPath()
        logging.info('Importing segment %s in %s' % (segment.getId(), filepath))
        tobiiglasses.utils.import_json_items_from_gzipfile(filepath, tobiiglasses.entities.TobiiSegment.livedata_filename, self.__decodeJSON__)

    def __init_datatypes__(self):
        self.__gazedata__[GazeData.Timestamp] = GazeItem(GazeData.Timestamp, np.dtype('float'))
        self.__gazedata__[GazeData.Gidx] = GazeItem(GazeData.Gidx, np.dtype('u4'))
        self.__gazedata__[GazeData.LoggedEvents] = GazeItem(GazeData.LoggedEvents, np.dtype(object))
        self.__gazedata__[GazeData.GazePositionX] = GazeItem(GazeData.GazePositionX, np.dtype('float'))
        self.__gazedata__[GazeData.GazePositionY] = GazeItem(GazeData.GazePositionY, np.dtype('float'))
        self.__gazedata__[GazeData.GazePixelX] = GazeItem(GazeData.GazePixelX, np.dtype('u4'))
        self.__gazedata__[GazeData.GazePixelY] = GazeItem(GazeData.GazePixelY, np.dtype('u4'))
        self.__gazedata__[GazeData.Gaze3DPositionX] = GazeItem(GazeData.Gaze3DPositionX, np.dtype('float'))
        self.__gazedata__[GazeData.Gaze3DPositionY] = GazeItem(GazeData.Gaze3DPositionY, np.dtype('float'))
        self.__gazedata__[GazeData.Gaze3DPositionZ] = GazeItem(GazeData.Gaze3DPositionZ, np.dtype('float'))
        self.__gazedata__[GazeData.Depth] = GazeItem(GazeData.Depth, np.dtype('float'))
        self.__gazedata__[GazeData.Vergence] = GazeItem(GazeData.Vergence, np.dtype('float'))
        self.__gazedata__[GazeData.Version] = GazeItem(GazeData.Version, np.dtype('float'))
        self.__gazedata__[GazeData.Tilt] = GazeItem(GazeData.Tilt, np.dtype('float'))
        self.__gazedata__[GazeData.GazeDirectionX_Left] = GazeItem(GazeData.GazeDirectionX_Left, np.dtype('float'))
        self.__gazedata__[GazeData.GazeDirectionY_Left] = GazeItem(GazeData.GazeDirectionY_Left, np.dtype('float'))
        self.__gazedata__[GazeData.GazeDirectionZ_Left] = GazeItem(GazeData.GazeDirectionZ_Left, np.dtype('float'))
        self.__gazedata__[GazeData.GazeDirectionX_Right] = GazeItem(GazeData.GazeDirectionX_Right, np.dtype('float'))
        self.__gazedata__[GazeData.GazeDirectionY_Right] = GazeItem(GazeData.GazeDirectionY_Right, np.dtype('float'))
        self.__gazedata__[GazeData.GazeDirectionZ_Right] = GazeItem(GazeData.GazeDirectionZ_Right, np.dtype('float'))
        self.__gazedata__[GazeData.PupilCenterX_Left] = GazeItem(GazeData.PupilCenterX_Left, np.dtype('float'))
        self.__gazedata__[GazeData.PupilCenterY_Left] = GazeItem(GazeData.PupilCenterY_Left, np.dtype('float'))
        self.__gazedata__[GazeData.PupilCenterZ_Left] = GazeItem(GazeData.PupilCenterZ_Left, np.dtype('float'))
        self.__gazedata__[GazeData.PupilCenterX_Right] = GazeItem(GazeData.PupilCenterX_Right, np.dtype('float'))
        self.__gazedata__[GazeData.PupilCenterY_Right] = GazeItem(GazeData.PupilCenterY_Right, np.dtype('float'))
        self.__gazedata__[GazeData.PupilCenterZ_Right] = GazeItem(GazeData.PupilCenterZ_Right, np.dtype('float'))
        self.__gazedata__[GazeData.PupilDiameter_Left] = GazeItem(GazeData.PupilDiameter_Left, np.dtype('float'))
        self.__gazedata__[GazeData.PupilDiameter_Right] = GazeItem(GazeData.PupilDiameter_Right, np.dtype('float'))

    def getData(self):
        return self.__gazedata__.values

    def getExpVarsHeaders(self):
        return self.__experiment_vars_headers__

    def getLoggedEvents(self):
        return self.__gazedata__[GazeData.LoggedEvents].getData()

    def getTimestamps(self):
        return self.__gazedata__[GazeData.Timestamp].values()

    def getVTS(self):
        return self.__vts__

    def toDataFrame(self):
        table = {}
        for label, data in self.__gazedata__.items():
            table[label] = pd.Series(data.getData(), dtype=object)
        return pd.DataFrame(table)

    def to_pickle(self, filename):
        self.toDataFrame().to_pickle(filename)



# Copyright (C) 2019  Davide De Tommaso
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from tobiiglasses.filters.models import DataFrameFilter
from tobiiglasses.gazedata import GazeData

class TimestampFilter(DataFrameFilter):
    def __init__(self, ts_list):
        DataFrameFilter.__init__(self)
        self.__ts_list__ = ts_list

    def __filter_condition__(self, df, tslist_to_exclude=[], columns=None):
        if len(tslist_to_exclude) > 0:
            res = list(set(self.__ts_list__) - set(tslist_to_exclude))
        else:
            res = self.__ts_list__
        return df[GazeData.Timestamp].isin(res)

    def getIntervalDuration(self):
        ts_list = self.getTimestamps()
        return (ts_list[-1] - ts_list[0])

    def getTimestamps(self):
        return self.__ts_list__


class SingleTimestamp(TimestampFilter):
    def __init__(self, ts):
        TimestampFilter.__init__(self, [ts])

class BetweenTimestamps(TimestampFilter):
    def __init__(self, t1, t2):
        TimestampFilter.__init__(self, list(range(int(t1),int(t2)+1)))

    def __filter_condition__(self, df, tslist_to_exclude=[], columns=None):
        return df[GazeData.Timestamp].between(self.__ts_list__[0], self.__ts_list__[-1])

class SingleLoggedEvent(DataFrameFilter):
    def __init__(self, logged_event):
        DataFrameFilter.__init__(self)
        self.__event__ = logged_event

    def __filter_condition__(self, df, tslist_to_exclude=[], columns=None):
        return df[GazeData.LoggedEvents] == self.__event__

class AroundLoggedEvents(DataFrameFilter):
    def __init__(self, logged_event, offset):
        DataFrameFilter.__init__(self)
        self.__logged_event__ = logged_event
        self.__offset__ = offset

    def __filter_condition__(self, df, tslist_to_exclude=[], columns=None):
        event_index = []
        f = SingleLoggedEvent(self.__logged_event__)
        res, ts_logged_events = f.getFilteredData(df, columns=[GazeData.Timestamp, GazeData.LoggedEvents])
        ts_list = []
        for i in range(0, len(ts_logged_events)):
            if i==len(ts_logged_events):
                break
            if self.__offset__ >= 0:
                t_from = df[GazeData.Timestamp] >= ts_logged_events[i]
                t_to = df[GazeData.Timestamp] <= ts_logged_events[i] + self.__offset__
            else:
                t_from = df[GazeData.Timestamp] >= ts_logged_events[i] + self.__offset__
                t_to = df[GazeData.Timestamp] <= ts_logged_events[i]

            a = df[t_from & t_to]
            ts_list.extend( list(a[GazeData.Timestamp]) )

        f = TimestampFilter(ts_list)
        return f.__filter_condition__(df, tslist_to_exclude, columns)


class BetweenLoggedEvents(DataFrameFilter):
    def __init__(self, from_loggedevent, to_loggedevent):
        DataFrameFilter.__init__(self)
        self.__from__ = from_loggedevent
        self.__to__ = to_loggedevent

    def __filter_condition__(self, df, tslist_to_exclude=[], columns=None):
        event_index = []
        f = SingleLoggedEvent(self.__from__)
        res, ts_from = f.getFilteredData(df, columns=[GazeData.Timestamp, GazeData.LoggedEvents])
        f = SingleLoggedEvent(self.__to__)
        res, ts_to = f.getFilteredData(df, columns=[GazeData.Timestamp, GazeData.LoggedEvents])
        ts_list = []
        for i in range(0, len(ts_from)):
            if i==len(ts_from) or i==len(ts_to):
                break
            t_from = df[GazeData.Timestamp] >= ts_from[i]
            t_to = df[GazeData.Timestamp] <= ts_to[i]
            a = df[t_from & t_to]
            ts_list.extend( list(a[GazeData.Timestamp]) )

        f = TimestampFilter(ts_list)
        return f.__filter_condition__(df, tslist_to_exclude, columns)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import json_utils
import datetime
import os

class Segment:
    video_filename = "fullstream.mp4"
    livedata_filename = "livedata.json.gz"
    segment_filename = "segment.json"

    def __init__(self, segment_dir, segment_id):
        self.segment_dir = segment_dir
        self.segment_id = segment_id
        json_utils.JSON_File(segment_dir, Segment.segment_filename).load(self.json_segment_hook)

    def json_segment_hook(self, item):
        self.seg_length_us = int(item["seg_length_us"])
        self.seg_calibrated = bool(item["seg_calibrated"])
        self.seg_t_start = datetime.datetime.strptime(item["seg_t_start"], '%Y-%m-%dT%H:%M:%S+%f')
        self.seg_t_stop = datetime.datetime.strptime(item["seg_t_stop"], '%Y-%m-%dT%H:%M:%S+%f')

    def getVideoFilepath(self):
        return os.path.join(self.segment_dir, Segment.video_filename)

    def getLiveDataFilepath(self):
        return os.path.join(self.segment_dir, Segment.livedata_filename)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import logging
import os
import csv
import gzip
import numpy as np
import pandas as pd
import cv2
import tobiiglasses.entities
from tobiiglasses.entities import TobiiRecording, TobiiSegment
from tobiiglasses.gazedata import GazeData
from tobiiglasses.events import GazeEvents
from tobiiglasses.exporter import RawCSV, ExtendedRawCSV

class Recording:

    def __init__(self, root_dir, project_id, recording_id):
        recordings_path = tobiiglasses.entities.get_recordings_path(root_dir, project_id)
        self.__recording__ = TobiiRecording(recordings_path, recording_id)
        self.__segment_data__ = {}
        self.__segment_events__ = {}
        self.__segment_ids__ = []
        self.__loadSegmentIDs__()

    def __exportData_CSV__(self, filepath=None, filename=None, csv_suffix='', segment_id=None, exporter=None):
        segment_ids = self.__getSegmentIDs__(segment_id)
        for s_id in segment_ids:
            gazedata = self.getGazeData(s_id)
            (filepath, filename) = self.__getFileParams__('csv', s_id, filepath, filename, csv_suffix)
            e = exporter(filepath, filename, gazedata)
            e.toCSV()

    def __getFileParams__(self, extension, segment_id=None, filepath=None, filename=None, suffix=''):
        if segment_id is None:
            segment_id = 1
        if filepath is None:
            filepath = "."
        if filename is None:
            filename = "%s-S%s-%s.%s" % (self.getParticipantName(), str(segment_id), suffix, extension)
        return (filepath, filename)

    def __getSegmentIDs__(self, segment_id=None):
        segment_ids = []
        if segment_id is None:
            segment_ids.extend(self.__segment_ids__)
        else:
            segment_ids.append(segment_id)
        return segment_ids

    def __loadSegmentIDs__(self):
        self.__segment_ids__.extend(range(1, self.__recording__.getSegmentsN() + 1))

    def exportCSV_ExtendedRawData(self, filepath=None, filename=None, segment_id=None):
        self.__exportData_CSV__(filepath, filename, 'eRaw', segment_id, ExtendedRawCSV)

    def exportCSV_Fixations(self, fixation_filter, filepath=None, filename=None, ts_filter=None, segment_id=None, aoi_model=None, aoi_plot=None):
        fixations = self.getFixations(fixation_filter, ts_filter, segment_id, aoi_model)
        (filepath, filename) = self.__getFileParams__('csv', segment_id, filepath, filename, 'Fixations')
        fixations.exportCSV(filepath, filename, ts_filter=ts_filter)
        if not aoi_model is None and aoi_plot is True:
            (filepath, filename) = self.__getFileParams__('pdf', segment_id, filepath, None, 'AOIs')
            aoi_model.savePlot(title=filename, filename=os.path.join(filepath, filename))

    def exportCSV_RawData(self, filepath=None, filename=None, segment_id=1):
        self.__exportData_CSV__(filepath, filename, 'Raw', segment_id, RawCSV)

    def getFixations(self, fixation_filter, ts_filter=None, segment_id=None, aoi_model=None):
        segment_ids = self.__getSegmentIDs__(segment_id)
        for s_id in segment_ids:
            gaze_data = self.getGazeData(s_id)
            gaze_events = self.getGazeEvents(s_id)
            gaze_events.filterFixations(fixation_filter, gaze_data.toDataFrame(), ts_filter, aoi_model)
            if not aoi_model is None:
                aoi_model.fit(gaze_events, ts_filter)
        return gaze_events

    def getGazeData(self, segment_id=1):
        segment = self.__recording__.getSegment(segment_id)
        if not segment_id in self.__segment_data__.keys():
            self.__segment_data__[segment_id] = GazeData(segment)
        return self.__segment_data__[segment_id]

    def getGazeEvents(self, segment_id):
        segment = self.__recording__.getSegment(segment_id)
        if not segment_id in self.__segment_events__.keys():
            self.__segment_events__[segment_id] = GazeEvents()
            events = self.getGazeData(segment_id).getLoggedEvents()
            for ts, value in events.items():
                self.__segment_events__[segment_id].addLoggedEvent(ts, value)
        return self.__segment_events__[segment_id]

    def getParticipantName(self):
        return self.__recording__.getParticipant().getName()

    def replay(self, segment_id, fps=25):
        data = self.getGazeData(segment_id)
        df = data.toDataFrame()
        vts = data.getVTS()
        vts_list = vts.keys()
        frame_duration = int(1000/fps)
        current_time = vts_list[0]
        f = self.__recording__.getSegment(segment_id).getVideoFilename()
        cap = cv2.VideoCapture(f)
        if (cap.isOpened() == False):
            print("Error opening video stream or file")
        i = 0
        while(cap.isOpened()):
            ret, frame = cap.read()
            if ret == True:
                if current_time > vts_list[i]:
                    i+=1
                delay = int(vts[vts_list[i]])
                T = df.index[df.index.get_loc(current_time+i, method='nearest')]
                x = df.at[T, GazeData.GazePixelX]
                y = df.at[T, GazeData.GazePixelY]
                if x>0 and y>0:
                    cv2.circle(frame,(int(x),int(y)), 30, (0,0,255), 2)
                cv2.imshow('Frame',frame)
                if cv2.waitKey(25) & 0xFF == ord('q'):
                    break
            else:
                break
            current_time+=frame_duration
        cap.release()
        cv2.destroyAllWindows()

    def saveVideoSnapshot(self, filename, ts, segment_id, fps=25):
        data = self.getGazeData(segment_id)
        df = data.toDataFrame()
        vts = data.getVTS()
        vts_list = vts.keys()
        frame_duration = int(1000/fps)
        current_time = vts_list[0]
        f = self.__recording__.getSegment(segment_id).getVideoFilename()
        cap = cv2.VideoCapture(f)
        if (cap.isOpened() == False):
            print("Error opening video stream or file")
        i = 0
        while(cap.isOpened()):
            ret, frame = cap.read()
            if ret == True:
                if current_time > vts_list[i]:
                    i+=1
                delay = int(vts[vts_list[i]])
                T = df.index[df.index.get_loc(current_time+i, method='nearest')]
                x = df.at[T, GazeData.GazePixelX]
                y = df.at[T, GazeData.GazePixelY]
                if T > ts:
                    cv2.imwrite(filename,frame)
                    break
            current_time+=frame_duration
        cap.release()



__all__ = [ "entities",
            "exporter",
            "filters",
            "gazedata",
            "livedata",
            "livedatajson",
            "metrics",
            "recordings",
            "segments",
            "utils"]



import json
import logging
import gzip
import os


def load_json_from_file(source_dir, filename):
    logging.info('Loading JSON from %s in %s' % (filename, source_dir))
    with open(os.path.join(source_dir, filename)) as f:
        try:
            res = json.load(f)
        except:
            res = None
    return res

def import_json_items_from_gzipfile(source_dir, filename, decode_JSON=None):
    logging.info('Importing JSON items from %s in %s' % (filename, source_dir))
    with gzip.open(os.path.join(source_dir, filename)) as f:
        for item in f:
            json.loads(item.decode('utf-8'), object_hook=decode_JSON)

def import_df_from_pickle(source_dir, filename):
    logging.info('Importing dataframe from %s in %s' % (filename, source_dir))
    return pd.read_pickle(os.path.join(source_dir, filename))



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

from tobiiglasses.filters.df import BetweenTimestamps
from tobiiglasses.events import GazeEvents
import numpy as np
import pandas as pd

class Stats:
    def __init__(self, df):
        self.__stats__ = df.describe()

    def __repr__(self):
        return str(self.__stats__)

    def quart25(self):
        return self.__stats__['25%']

    def quart50(self):
        return self.__stats__['50%']

    def quart75(self):
        return self.__stats__['75%']

    def count(self):
        return self.__stats__['count']

    def max(self):
        return self.__stats__['max']

    def mean(self):
        return self.__stats__['mean']

    def median(self):
        return self.quart50()

    def min(self):
        return self.__stats__['min']

    def std(self):
        return self.__stats__['std']


class Fixations_Metrics:
    def __init__(self, gaze_events):
        self.__events__ = gaze_events
        self.__fixations_df__ = gaze_events.getFixations()

    def __filter__(self, ts_filter):
        if ts_filter is None:
            return (self.__fixations_df__, self.__events__.getTimestamps())
        return ts_filter.getFilteredData(self.__fixations_df__)

    def getFixationsCount(self, ts_filter=None):
        df, ts_list = self.__filter__(ts_filter)
        return len(df.values)

    def getFixationDurationStats(self, ts_filter=None):
        df, ts_list = self.__filter__(ts_filter)
        return Stats(pd.to_numeric(df[GazeEvents.EventDuration]))

    def getFixationsXStats(self, ts_filter=None):
        df, ts_list = self.__filter__(ts_filter)
        return Stats(pd.to_numeric(df[GazeEvents.Fixation_X]))

    def getFixationsYStats(self, ts_filter=None):
        df, ts_list = self.__filter__(ts_filter)
        return Stats(pd.to_numeric(df[GazeEvents.Fixation_X]))

    def getAOIs_TTFF(self, ts_onset=0):
        ts_filter = BetweenTimestamps(t1=ts_onset, t2=self.__events__.getTimestamps()[-1])
        df, ts_list = self.__filter__(ts_filter)
        aois = df.AOI.unique()
        aois_ttff = {}
        for item in aois:
            res = df.loc[df.AOI == item, GazeEvents.Timestamp]
            if res.count() > 0:
                aois_ttff[item] = res.iloc[0] - ts_list[0]
            else:
                aois_ttff[item] = None
        return aois_ttff

    def getAOIs_FirstFixationDuration(self, ts_onset=0):
        ts_filter = BetweenTimestamps(t1=ts_onset, t2=self.__events__.getTimestamps()[-1])
        df, ts_list = self.__filter__(ts_filter)
        aois = df.AOI.unique()
        aois_ffd = {}
        for item in aois:
            res = df.loc[df.AOI == item, GazeEvents.EventDuration]
            if res.count() > 0:
                aois_ffd[item] = res.iloc[0]
            else:
                aois_ffd[item] = None
        return aois_ffd

    def getAOIs_TotalFixationDuration(self, ts_filter=None):
        df, ts_list = self.__filter__(ts_filter)
        aois = df.AOI.unique()
        aois_tfd = {}
        tot = df.loc[:, GazeEvents.EventDuration].sum()
        for item in aois:
            aoi_sum = df.loc[df.AOI == item, GazeEvents.EventDuration].sum()
            if aoi_sum > 0:
                aois_tfd[item] = (aoi_sum*100.0)/tot
            else:
                aois_tfd[item] = 0
        return aois_tfd



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='tobiiglasses',
    version='0.5',
    description='A Python suite for Tobii Pro Glasses 2',
    url='https://github.com/ddetommaso/TobiiGlassesPySuite',
    download_url='https://github.com/ddetommaso/TobiiGlassesPySuite/archive/0.5.tar.gz',
    install_requires=['tobiiglassesctrl>=2.2.1'],
    author='Davide De Tommaso',
    author_email='ing.davidedetommaso@gmail.com',
    keywords=['eye-tracker','tobii','glasses', 'tobii pro glasses 2', 'tobii glasses', 'eye tracking'],
    packages=find_packages(exclude=['examples*']),
    classifiers = [
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.5'
    ],
)



# Copyright (C) 2019  Davide De Tommaso
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import pandas as pd
import math
from sortedcontainers import SortedList, SortedDict
from tobiiglasses.gazedata import GazeData

class FixationsFilter:

    def __init__(self):
        self.__x__ = None
        self.__y__ = None
        self.__fixation_index__ = 0
        self.__saccade_index__ = 0

    def filter(self, gaze_events):
        pass

    def setData(self, x, y):
        self.__x__ = x.astype('float')
        self.__y__ = y.astype('float')
        self.__validation_check__()

    def __addFixation__(self, ts, duration, fixation_x, fixation_y, gaze_events):
        if math.isnan(fixation_x) or math.isnan(fixation_y) or math.isinf(fixation_x) or math.isinf(fixation_y):
            pass
        else:
            gaze_events.addFixation(ts, self.__fixation_index__, int(duration), int(fixation_x), int(fixation_y))
            self.__fixation_index__+=1

    def __addSaccade__(self, ts, duration, saccade_start_x, saccade_start_y, saccade_end_x, saccade_end_y, gaze_events):
        gaze_events.addSaccade(ts, self.__saccade_index__, duration, saccade_start_x, saccade_start_y, saccade_end_x, saccade_end_y)
        self.__saccade_index__+=1

    def __validation_check__(self):
        if self.__x__ is None or self.__y__ is None:
            raise ValueError('The FixationFilter requires to set first x and y as pandas Series, please verify the correct use of setData()')
        if not isinstance(self.__x__, pd.core.series.Series):
            raise ValueError('The FixationFilter requires (x) variables as pandas Series')
        elif not isinstance(self.__y__, pd.core.series.Series):
            raise ValueError('The FixationFilter requires (y) variables as pandas Series')
        if len(self.__x__) != len(self.__y__):
            raise ValueError('The FixationFilter needs variables (x) and (y) with the same size')


class DataFrameFilter(object):

    def __init__(self):
        pass

    def __filter_condition__(self, df, tslist_to_exclude=[], columns=None):
        raise NotImplementedError( "DataFrameFilter should have implemented a filter condition" )

    def getFilteredData(self, df, tslist_to_exclude=[], columns=None):
        if columns is None:
            columns = list(df.columns.values)
        res = df.filter(items=columns)
        df_filtered = res[self.__filter_condition__(res, tslist_to_exclude, columns)]
        return (df_filtered, df_filtered[GazeData.Timestamp].values)



#   Copyright (C) 2019  Davide De Tommaso
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>

import pandas as pd
import numpy as np
import tobiiglasses
import logging
import os
from sortedcontainers import SortedList, SortedDict
from tobiiglasses.gazedata import GazeItem, GazeData
from tobiiglasses.filters.df import TimestampFilter
from tobiiglasses.exporter import FixationsCSV


class GazeEvents:
    Timestamp = "Timestamp"
    LoggedEvents = "Logged Events"
    GazeType = "Gaze Type"
    EventIndex = "Event Index"
    EventDuration = "Event Duration"
    Fixation_X = "Fixation X"
    Fixation_Y = "Fixation Y"
    AOI = "AOI"
    AOI_Distance = "AOI Distance"
    Saccade_Start_X = "Saccade Start X"
    Saccade_Start_Y = "Saccade Start Y"
    Saccade_End_X = "Saccade End X"
    Saccade_End_Y = "Saccade End Y"

    InputFixationFilterX = GazeData.GazePixelX
    InputFixationFilterY = GazeData.GazePixelY

    def __init__(self):
        self.__events__ = SortedDict({})
        self.__init_datatypes__()
        self.__fixation_index__ = 0
        self.__saccade_index__ = 0
        self.__ts_processed__ = []

    def __getitem__(self, key):
        return self.__events__[key]

    def __getFilteredGazeData__(self, gazedata_df, ts_filter=None):
        if ts_filter is None:
            ts_filter = TimestampFilter(list(gazedata_df[GazeData.Timestamp]))
        df, ts_list = ts_filter.getFilteredData(gazedata_df, self.__ts_processed__)
        self.__ts_processed__.extend(ts_list)
        return df, ts_list

    def __init_datatypes__(self):
        self.__events__[GazeEvents.Timestamp] = GazeItem(GazeEvents.Timestamp, np.dtype('float'))
        self.__events__[GazeEvents.LoggedEvents] = GazeItem(GazeEvents.LoggedEvents, np.dtype(object))
        self.__events__[GazeEvents.GazeType] = GazeItem(GazeEvents.GazeType, np.dtype(object))
        self.__events__[GazeEvents.Fixation_X] = GazeItem(GazeEvents.Fixation_X, np.dtype('u4'))
        self.__events__[GazeEvents.Fixation_Y] = GazeItem(GazeEvents.Fixation_Y, np.dtype('u4'))
        self.__events__[GazeEvents.EventIndex] = GazeItem(GazeEvents.EventIndex, np.dtype('u4'))
        self.__events__[GazeEvents.EventDuration] = GazeItem(GazeEvents.EventDuration, np.dtype('u4'))
        self.__events__[GazeEvents.AOI] = GazeItem(GazeEvents.AOI, np.dtype(object))
        self.__events__[GazeEvents.AOI_Distance] = GazeItem(GazeEvents.AOI_Distance, np.dtype('f2'))
        self.__events__[GazeEvents.Saccade_Start_X] = GazeItem(GazeEvents.Saccade_Start_X, np.dtype('f2'))
        self.__events__[GazeEvents.Saccade_Start_Y] = GazeItem(GazeEvents.Saccade_Start_Y, np.dtype('f2'))
        self.__events__[GazeEvents.Saccade_End_X] = GazeItem(GazeEvents.Saccade_End_X, np.dtype('f2'))
        self.__events__[GazeEvents.Saccade_End_Y] = GazeItem(GazeEvents.Saccade_End_Y, np.dtype('f2'))

    def addFixation(self, ts, index, duration, fixation_x, fixation_y, aoi=None):
        self.__events__[GazeEvents.Timestamp][ts] = ts
        self.__events__[GazeEvents.GazeType][ts] = "Fixation"
        self.__events__[GazeEvents.EventIndex][ts] = index
        self.__events__[GazeEvents.EventDuration][ts] = duration
        self.__events__[GazeEvents.Fixation_X][ts] = fixation_x
        self.__events__[GazeEvents.Fixation_Y][ts] = fixation_y

    def addLoggedEvent(self, ts, logged_event):
        self.__events__[GazeEvents.Timestamp][ts] = ts
        self.__events__[GazeEvents.LoggedEvents][ts] = logged_event

    def addSaccade(self, ts, index, duration, saccade_start_x, saccade_start_y, saccade_end_x, saccade_end_y):
        self.__events__[GazeEvents.Timestamp][ts] = ts
        self.__events__[GazeEvents.GazeType][ts] = "Saccade"
        self.__events__[GazeEvents.EventIndex][ts] = index
        self.__events__[GazeEvents.EventDuration][ts] = duration
        self.__events__[GazeEvents.Saccade_Start_X][ts] = saccade_start_x
        self.__events__[GazeEvents.Saccade_Start_Y][ts] = saccade_start_y
        self.__events__[GazeEvents.Saccade_End_X][ts] = saccade_end_x
        self.__events__[GazeEvents.Saccade_End_Y][ts] = saccade_end_y

    def exportCSV(self, filepath, filename, ts_filter=None):
        fixations_df = self.toDataFrame(ts_filter).dropna(subset=[GazeEvents.Fixation_X, GazeEvents.Fixation_Y])
        exp = FixationsCSV(filepath, filename, fixations_df)
        exp.toCSV()

    def exportDF(self, filepath, filename, ts_filter=None):
        logging.info('Exporting gaze events in %s' % filename)
        path = os.path.join(filepath, filename)
        self.toDataFrame(ts_filter).dropna(subset=[GazeEvents.Fixation_X, GazeEvents.Fixation_Y]).to_pickle(path)

    def filterFixations(self, fixation_filter, gazedata_df, ts_filter=None, aoi_model=None):
        df, ts_list = self.__getFilteredGazeData__(gazedata_df, ts_filter)
        if not df.empty:
            x = pd.Series(df[GazeEvents.InputFixationFilterX])
            y = pd.Series(df[GazeEvents.InputFixationFilterY])
            fixation_filter.setData(x,y)
            fixation_filter.filter(self)

    def getFixations(self, ts_filter=None):
        df = self.toDataFrame(ts_filter)
        return df.loc[df[GazeEvents.GazeType] == 'Fixation']

    def getSaccades(self, ts_filter=None):
        df = self.toDataFrame(ts_filter)
        return df.loc[df[GazeEvents.GazeType] == 'Saccade']

    def getFixationsAsNumpy(self, ts_filter):
        fixations_df = self.toDataFrame(ts_filter).dropna(subset=[GazeEvents.Fixation_X, GazeEvents.Fixation_Y])
        x = fixations_df[GazeEvents.Fixation_X].values
        y = fixations_df[GazeEvents.Fixation_Y].values
        ts_list = fixations_df[GazeEvents.Timestamp].values.tolist()
        return (ts_list, x, y)

    def getTimestamps(self):
        return list(self.__events__[GazeEvents.Timestamp].values())

    def setAOI(self, ts, aoi_label, aoi_distance):
        self.__events__[GazeEvents.AOI][ts] = aoi_label
        self.__events__[GazeEvents.AOI_Distance][ts] = aoi_distance

    def toDataFrame(self, ts_filter=None):
        table = {}
        for label, data in self.__events__.items():
            table[label] = pd.Series(data.getData(), dtype=object)
        df = pd.DataFrame(table)
        if ts_filter is None:
            return df
        else:
            filtered_df, ts_list = ts_filter.getFilteredData(df)
            return filtered_df

    def to_pickle(self, filename, ts_filter=None):
        self.toDataFrame(ts_filter).to_pickle(filename)



# Copyright (C) 2019  Davide De Tommaso
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>
#


# Implementation of the I-DT algorithm as described in:
#
# Dario D. Salvucci and Joseph H. Goldberg. 2000. Identifying fixations and
# saccades in eye-tracking protocols. In Proceedings of the 2000 symposium
# on Eye tracking research & applications (ETRA '00). ACM, New York, NY, USA,
# 71-78. DOI=http://dx.doi.org/10.1145/355017.355028

import pandas as pd
import numpy as np
import logging
import math
from tobiiglasses.filters.models import FixationsFilter

FIXATION_MAX_DURATION = 1000

class FilterDT(FixationsFilter):
    def __init__(self, dispersion_threshold=10, duration_threshold=100):
        FixationsFilter.__init__(self)
        self.__dispersion_threshold__ = dispersion_threshold
        self.__duration_threshold__ = duration_threshold

    def __init_time_window__(self, ts, start_index):
        ts_list = []
        for stop_index in range(start_index, len(ts)):
            ts_list.append(ts[stop_index])
            if ts[stop_index]-ts[start_index] >= self.__duration_threshold__:
                break
        return (ts_list, stop_index)

    def __getCentroid__(self, x, y):
        return (np.mean(x), np.mean(y))

    def __getEuclideanDispersion__(self, x, y):
        cx, cy = self.__getCentroid__(x, y)
        c = [cx, cy]
        points = np.column_stack([x, y])
        dist = (points - c)**2
        dist = np.sum(dist, axis=1)
        dist = np.sqrt(dist)
        return max(dist)

    # dispersion defined as in the paper
    def __getDispersion__(self, x, y):
        return (max(x) - min(x)) + (max(y) - min(y))

    def filter(self, gaze_events):
        logging.info("FilterDT is processing gaze data ...")
        n_samples = self.__x__.size
        fixations_x = []
        fixations_y = []
        fixations_duration = []
        fixations_index = []
        start_index = 0
        ts = self.__x__.keys()
        while start_index < n_samples-1:
            (ts_list, stop_index) = self.__init_time_window__(ts, start_index)
            d = self.__getEuclideanDispersion__( self.__x__[ts_list], self.__y__[ts_list])
            if d <= self.__dispersion_threshold__:
                while True:
                    d = self.__getEuclideanDispersion__(self.__x__[ts_list], self.__y__[ts_list])
                    if stop_index == n_samples - 1:
                        break
                    elif d > self.__dispersion_threshold__:
                        break
                    else:
                        stop_index += 1
                        ts_list.append(ts[stop_index])
                ts_list.pop(-1)
                cx,cy = self.__getCentroid__(self.__x__[ts_list], self.__y__[ts_list])
                dur = ts_list[-1] - ts_list[0]
                if dur > FIXATION_MAX_DURATION:
                    dur = FIXATION_MAX_DURATION
                self.__addFixation__(ts_list[0], dur, cx, cy, gaze_events)
                start_index = stop_index
            else:
                start_index += 1
